from celery.utils.log import get_task_logger

from firexapp.broker_manager.broker_factory import BrokerFactory
from firexapp.plugins import get_plugin_module_list
from firexapp.discovery import find_firex_task_bundles
logger = get_task_logger(__name__)


broker_url = BrokerFactory.get_broker_url()
result_backend = broker_url

# find default tasks
logger.debug("Beginning bundle discovery")
bundles = find_firex_task_bundles()
logger.debug("Bundle discovery completed")

imports = tuple(bundles) + tuple(get_plugin_module_list()) + tuple(["firexapp.tasks.example",
                                                                    "firexapp.tasks.core_tasks",
                                                                    "firexapp.submit.report_trigger"])

root_task = "firexapp.tasks.core_tasks.RootTask"

accept_content = ['pickle']
task_serializer = 'pickle'
result_serializer = 'pickle'

result_expires = None

task_acks_late = True
worker_prefetch_multiplier = 1



import time

import os
import psutil
import re
import socket


FIREX_BIN_DIR_ENV = 'firex_bin_dir'


def delimit2list(str_to_split, delimiters=(',', ';', '|', ' ')) -> []:
    if not str_to_split:
        return []

    if not isinstance(str_to_split, str):
        return str_to_split

    # regex for only comma is (([^,'"]|"(?:\\.|[^"])*"|'(?:\\.|[^'])*')+)
    regex = """(([^""" + "".join(delimiters).replace(" ", "\s") + """'"]|"(?:\\.|[^"])*"|'(?:\\.|[^'])*')+)"""
    tokens = re.findall(regex, str_to_split)

    # unquote "tokens" if necessary
    tokens = [g1 if g1.strip() != g2.strip() or g2[0] not in "'\"" else g2.strip(g2[0]) for g1, g2 in tokens]
    tokens = [t.strip() for t in tokens]  # remove extra whitespaces
    tokens = [t.strip("".join(delimiters) + " ") for t in tokens]  # remove any extra (or lone) delimiters
    tokens = [t for t in tokens if t]  # remove empty tokens
    return tokens


def get_available_port(so_reuseport=True):
    sock = socket.socket()
    if so_reuseport and hasattr(socket, 'SO_REUSEPORT'):
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(('', 0))
    port = sock.getsockname()[1]
    return port


def silent_mkdir(path, mode=0o777, exist_ok=True):
    os.makedirs(path, mode=mode, exist_ok=exist_ok)


def poll_until_file_exist(file_path, timeout=10):
    timeout_time = time.time() + timeout
    while not os.path.exists(file_path) and time.time() < timeout_time:
        time.sleep(0.1)
    assert os.path.isfile(file_path), 'File %s did not exist within %r seconds' % (file_path, timeout)


def poll_until_existing_file_not_empty(file_path, timeout=10):
    timeout_time = time.time() + timeout
    while os.stat(file_path).st_size == 0 and time.time() < timeout_time:
        time.sleep(0.1)
    assert os.stat(file_path).st_size > 0, 'File %s size is zero' % file_path


def poll_until_file_not_empty(file_path, timeout=10):
    start_time = time.time()
    poll_until_file_exist(file_path, timeout)
    remaining_timeout = timeout - (time.time() - start_time)
    poll_until_existing_file_not_empty(file_path, remaining_timeout)


def poll_until_dir_empty(dir_path, timeout=15):
    timeout_time = time.time() + timeout
    while len(os.listdir(dir_path)) > 0 and time.time() < timeout_time:
        time.sleep(0.1)
    return not os.listdir(dir_path)


def proc_matches(proc_info, pname, cmdline_regex, cmdline_contains):
    if proc_info['name'] == pname:
        if cmdline_regex:
            return any(cmdline_regex.search(item) for item in proc_info['cmdline'])
        elif cmdline_contains:
            return any(cmdline_contains in item for item in proc_info['cmdline'])
        else:
            return True
    else:
        return False


def find_procs(name, cmdline_regex=None, cmdline_contains=None):
    matching_procs = []
    if cmdline_regex:
        cmdline_regex = re.compile(cmdline_regex)
    else:
        cmdline_regex = None
    for proc in psutil.process_iter():
        try:
            pinfo = proc.as_dict(attrs=['name', 'cmdline', 'pid'])
        except psutil.NoSuchProcess:
            pass
        else:
            if proc_matches(pinfo, name, cmdline_regex, cmdline_contains):
                matching_procs.append(proc)

    return matching_procs


def wait_until(predicate, timeout, sleep_for, *args, **kwargs):
    max_time = time.time() + timeout
    while time.time() < max_time:
        if predicate(*args, **kwargs):
            return True
        time.sleep(sleep_for)
    return predicate(*args, **kwargs)


def wait_until_pid_not_exist(pid, timeout=7, sleep_for=1):
    return wait_until(lambda p: not psutil.pid_exists(p), timeout, sleep_for, pid)


def qualify_firex_bin(bin_name):
    if FIREX_BIN_DIR_ENV in os.environ:
        return os.path.join(os.environ[FIREX_BIN_DIR_ENV], bin_name)
    return bin_name


def select_env_vars(env_names):
    return {k: v for k, v in os.environ.items() if k in env_names}



import inspect
from celery.exceptions import NotRegistered
from firexapp.plugins import plugin_support_parser
from firexapp.application import import_microservices, get_app_task


class InfoBaseApp:
    def __init__(self):
        self._list_sub_parser = None
        self._info_sub_parser = None

    def create_list_sub_parser(self, sub_parser):
        list_parser = sub_parser.add_parser("list", help="Lists FireX microservices, or used arguments"
                                                         "  {microservices,arguments}",
                                            parents=[plugin_support_parser])
        list_group = list_parser.add_mutually_exclusive_group(required=True)
        list_group.add_argument("--microservices", '-microservices', help="Lists all available microservices",
                                action='store_true')
        list_group.add_argument("--arguments", '-arguments', help="Lists all arguments used by microservices",
                                action='store_true')

        list_group.set_defaults(func=self.run_list)

        return list_group

    def create_info_sub_parser(self, sub_parser):
        if not self._info_sub_parser:

            info_parser = sub_parser.add_parser("info", help="Lists detailed information about a microservice",
                                                parents=[plugin_support_parser])
            info_parser.add_argument("entity", help="The short or long name of the microservice to be detailed, or a "
                                                    "microservice argument")

            info_parser.set_defaults(func=self.run_info)
            self._info_sub_parser = info_parser
        return self._info_sub_parser

    def run_list(self, args):
        if args.microservices:
            self.print_available_microservices(args.plugins)
        elif args.arguments:
            self.print_argument_used(args.plugins)

    def run_info(self, args):
        self.print_details(args.entity, args.plugins)

    @staticmethod
    def print_available_microservices(plugins: str):
        apps = import_microservices(plugins)
        print()
        print("The following microservices are available:")

        services = [str(task) for task in apps]
        services = [task for task in services if not task.startswith('celery.')]  # filter out base celery types
        services.sort()
        for service in services:
            print(service)

        pointers = [(full, apps[full].name) for full in apps if apps[full].name not in full]
        if pointers:
            print("\nPointers (override -> original):")
            for new, old in pointers:
                print(new, "->", old)
        print("\nUse the info sub-command for more details\n")

    @staticmethod
    def print_argument_used(plugins: str):
        all_tasks = import_microservices(plugins)
        print()
        print("The following arguments are used by microservices:")
        usage = get_argument_use(all_tasks)
        for arg in sorted(usage):
            print(arg)
        print("\nUse the info sub-command for more details\n")

    def print_details(self, entity, plugins, all_tasks=None):
        if not all_tasks:
            all_tasks = import_microservices(plugins)

        # Is this thing a microservice?
        task = None
        try:
            task = get_app_task(entity, all_tasks)
        except NotRegistered:
            if "." in entity:
                try:
                    task = get_app_task(entity.split('.')[-1], all_tasks)
                except NotRegistered:
                    pass
        if task:
            self.print_task_details(task)
            return

        # Is this thing an argument
        all_args = get_argument_use(all_tasks)
        if entity in all_args:
            print("Argument name: " + entity)
            print("Used in the following microservices:")
            for micro in all_args[entity]:
                print(micro.name)
            return

        self._info_sub_parser.exit(status=-1, message="Microservice %s was not found!" % entity)

    @staticmethod
    def print_task_details(task):
        dash_length = 40
        print('-' * dash_length)
        split_name = task.name.split(".")
        name = split_name[-1]
        path = '.'.join(split_name[0:-1])
        if path:
            path = " (%s)" % path
        print("Name: " + name + path)

        if task.__doc__:
            print('\n' + inspect.getdoc(task))

        tab = ' ' * 1
        print("\nArguments info")
        print(  "--------------")
        print(" Mandatory:")
        required_args = getattr(task, "required_args", [])
        cnt = 0
        for chain_arg in sorted(required_args):
            if "self" not in chain_arg and \
               "uid" not in chain_arg and \
               chain_arg is not 'kwargs':
                print(tab, chain_arg)
                cnt += 1
        if not cnt:
            print(tab, "None")

        print("\n Optional:")
        optional_args = getattr(task, "optional_args", {})
        if len(optional_args):
            for chain_arg in sorted(optional_args):
                print(tab, chain_arg + "=" + str(optional_args[chain_arg]))
        else:
            print(tab, "None")

        print('\n Returns:')
        out = getattr(task, "return_keys", {})
        if out:
            for chain_arg in sorted(out):
                print(tab, chain_arg)
        else:
            print(tab, "None")

def get_argument_use(all_tasks) -> dict:
    argument_usage = {}
    for _, task in all_tasks.items():
        if not hasattr(task, "required_args") or not hasattr(task, "optional_args"):
            continue

        for arg in task.required_args + list(task.optional_args):
            if arg in argument_usage:
                argument_usage[arg].add(task)
            else:
                tasks = set()
                tasks.add(task)
                argument_usage[arg] = tasks

    return argument_usage



import json
import os
import time
import shlex
import subprocess
from firexapp.fileregistry import FileRegistry
from firexapp.submit.uid import Uid
from socket import gethostname
from urllib.parse import urlsplit
from logging import ERROR, INFO
from psutil import Process
from pathlib import Path

from firexapp.broker_manager import BrokerManager
from firexapp.common import get_available_port

REDIS_DIR_REGISTRY_KEY = 'REDIS_DIR_REGISTRY_KEY'
FileRegistry().register_file(REDIS_DIR_REGISTRY_KEY, os.path.join(Uid.debug_dirname, 'redis'))

REDIS_LOG_REGISTRY_KEY = 'REDIS_LOG_REGISTRY_KEY'
FileRegistry().register_file(REDIS_LOG_REGISTRY_KEY,
                             os.path.join(FileRegistry().get_relative_path(REDIS_DIR_REGISTRY_KEY), 'redis.stdout'))

REDIS_PID_REGISTRY_KEY = 'REDIS_PID_REGISTRY_KEY'
FileRegistry().register_file(REDIS_PID_REGISTRY_KEY,
                             os.path.join(FileRegistry().get_relative_path(REDIS_DIR_REGISTRY_KEY), 'redis.pid'))

REDIS_METADATA_REGISTRY_KEY = 'REDIS_METDATA_REGISTRY_KEY'
FileRegistry().register_file(REDIS_METADATA_REGISTRY_KEY,
                             os.path.join(FileRegistry().get_relative_path(REDIS_DIR_REGISTRY_KEY),
                                          'run-metadata.json'))


class RedisDidNotBecomeActive(Exception):
    pass


class RedisPortNotAssigned(Exception):
    pass


class RedisManager(BrokerManager):

    METADATA_BROKER_URL_KEY = 'broker_url'

    def __init__(self, redis_bin_base, hostname=gethostname(), port=None, logs_dir=None):
        self.redis_bin_base = redis_bin_base
        self.host = hostname
        self.port = port
        self.logs_dir = logs_dir

        self._log_file = None
        self._pid_file = None
        self._metadata_file = None

    @property
    def redis_cli_cmd(self):
        return self.get_redis_cli_cmd(self.port)

    def get_redis_cli_cmd(self, port):
        cmd = os.path.join(self.redis_bin_base, 'redis-cli') + ' -p %d' % port
        if self.host != gethostname():
            cmd += ' -h %s' % self.host
        return cmd

    @property
    def redis_server_cmd(self):
        return self.get_redis_server_cmd(self.port)

    def get_redis_server_cmd(self, port):
        return os.path.join(self.redis_bin_base, 'redis-server') + ' --port %d' % port

    @property
    def broker_url(self):
        return self.get_broker_url(self.port, self.host)

    @property
    def port(self):
        if not self.__port:
            raise RedisPortNotAssigned()
        return self.__port

    @port.setter
    def port(self, port):
        self.__port = int(port) if port else port

    @staticmethod
    def get_log_file(logs_dir):
        return FileRegistry().get_file(REDIS_LOG_REGISTRY_KEY, logs_dir)

    @staticmethod
    def get_pid_file(logs_dir):
        return FileRegistry().get_file(REDIS_PID_REGISTRY_KEY, logs_dir)

    @staticmethod
    def get_metdata_file(logs_dir):
        return FileRegistry().get_file(REDIS_METADATA_REGISTRY_KEY, logs_dir)

    @classmethod
    def read_metadata(cls, logs_dir):
        with open(cls.get_metdata_file(logs_dir)) as fp:
            metadata = json.load(fp)
        return metadata

    @classmethod
    def get_broker_url_from_metadata(cls, logs_dir):
        metadata = cls.read_metadata(logs_dir)
        return metadata[cls.METADATA_BROKER_URL_KEY]

    @property
    def log_file(self):
        if not self._log_file and self.logs_dir:
            _log_file = self.get_log_file(self.logs_dir)
            os.makedirs(os.path.dirname(_log_file), exist_ok=True)
            self._log_file = _log_file
        return self._log_file

    @property
    def pid_file(self):
        if not self._pid_file and self.logs_dir:
            _pid_file = self.get_pid_file(self.logs_dir)
            os.makedirs(os.path.dirname(_pid_file), exist_ok=True)
            self._pid_file = _pid_file
        return self._pid_file

    @property
    def metadata_file(self):
        if not self._metadata_file and self.logs_dir:
            _metadata_file = self.get_metdata_file(self.logs_dir)
            os.makedirs(os.path.dirname(_metadata_file), exist_ok=True)
            self._metadata_file = _metadata_file
        return self._metadata_file

    def create_metadata_file(self):
        if self.metadata_file:
            self.log('Creating %s' % self.metadata_file)
            data = {self.METADATA_BROKER_URL_KEY: self.broker_url}
            with open(self.metadata_file, 'w') as f:
                json.dump(data, f, sort_keys=True, indent=2)

    def _start(self):
        try:
            port = self.port
        except RedisPortNotAssigned:
            port = get_available_port()
        self.log('Starting new process (port %d)...' % port)
        cmd = self.get_redis_server_cmd(port) + ' --loglevel debug --protected-mode no --daemonize yes --timeout 0'
        if self.pid_file:
            cmd += ' --pidfile %s' % self.pid_file
        if self.log_file:
            cmd += ' --logfile %s' % self.log_file
        subprocess.check_call(shlex.split(cmd))
        self.wait_until_active(port=port)
        self.port = port
        self.create_metadata_file()
        self.log('redis started.')

    def start(self, max_retries=3):
        max_trials = max_retries + 1
        trials = 0

        while True:
            trials += 1
            try:
                self._start()
            except (subprocess.CalledProcessError, RedisDidNotBecomeActive):
                if trials >= max_trials:
                    self.log('Redis did not come up after %d trial(s) (max_trials=%d)..Giving up!' %
                             (trials, max_trials), level=ERROR)
                    raise
                self.log('Redis did not come up after %d trial(s) (max_trials=%d)' % (trials, max_trials), level=INFO)
            else:
                break

    def shutdown(self):
        try:
            RedisManager.log('shutting down...')
            self.cli('shutdown')
        except subprocess.CalledProcessError:
            RedisManager.log('could not shutdown.')

    def force_kill(self):
        try:
            RedisManager.log('force killing...')
            Process(int(Path(self.pid_file).read_text())).kill()
        except Exception:
            RedisManager.log('could not force kill.')

    def get_url(self) -> str:
        return self.broker_url

    def is_alive(self, port=None):
        port = self.port if port is None else port
        try:
            output = self.cli('PING', port=port)
        except subprocess.CalledProcessError:
            return False
        else:
            return output == 'PONG'

    def wait_until_active(self, timeout=60, port=None):
        port = self.port if port is None else port
        timeout_time = time.time() + timeout
        while time.time() < timeout_time:
            if self.is_alive(port=port):
                return
            time.sleep(0.1)
        raise RedisDidNotBecomeActive('Redis Server did not respond after %r seconds' % timeout)

    @staticmethod
    def get_broker_url(port=6379, hostname=gethostname()):
        return 'redis://%s:%d/0' % (hostname, int(port))

    @staticmethod
    def get_hostname_port_from_url(broker_url):
        hostname, port = urlsplit(broker_url).netloc.split(':')
        return hostname, port

    def get(self, key):
        return self.cli('GET %s' % key)

    def set(self, key, value):
        rc = self.cli('SET %s %s' % (key, value))
        assert rc == 'OK', 'The return value was %s' % rc

    def monitor(self, monitor_file):
        cmd = os.path.join(self.redis_bin_base, 'redis-cli') + ' -p %d MONITOR &' % self.port
        with open(monitor_file, 'w') as out:
            subprocess.check_call(cmd, shell=True, stdout=out, stderr=subprocess.STDOUT)

    def cli(self, cmd, port=None):
        port = self.port if port is None else port
        null = open(os.devnull, 'w')
        cmd = self.get_redis_cli_cmd(port=port) + ' %s' % cmd
        return subprocess.check_output(shlex.split(cmd), stderr=null).decode().strip()

    def __repr__(self):
        return self.broker_url

    def __eq__(self, other):
        return str(other) == self.broker_url



import os
from firexapp.broker_manager import BrokerManager
from firexapp.broker_manager.redis_manager import RedisManager
from firexapp.engine.celery import app

REDIS_BIN_ENV = "redis_bin_dir"

def get_redis_bin_dir():
    redis_bin_dir = os.environ.get(REDIS_BIN_ENV, "")
    if not redis_bin_dir:
        try:
            redis_bin_dir = app.conf.redis_bin_dir
        except AttributeError:
            pass
    return redis_bin_dir


class BrokerFactory:
    broker_env_variable = 'BROKER'

    @classmethod
    def set_broker_env(cls, broker_url):
        os.environ[cls.broker_env_variable] = broker_url

    @classmethod
    def create_new_broker_manager(cls, *args, **kwargs) -> BrokerManager:
        return RedisManager(*args, redis_bin_base=get_redis_bin_dir(), **kwargs)

    @classmethod
    def broker_manager_from_env(cls) -> BrokerManager:
        existing_broker_url = cls.get_broker_url(assert_if_not_set=True)
        hostname, port = RedisManager.get_hostname_port_from_url(existing_broker_url)
        return RedisManager(hostname=hostname,
                            port=port,
                            redis_bin_base=get_redis_bin_dir())

    @classmethod
    def get_broker_url_from_logs_dir(cls, logs_dir) -> str:
        return RedisManager.get_broker_url_from_metadata(logs_dir)

    @classmethod
    def broker_manager_from_logs_dir(cls, logs_dir) -> BrokerManager:
        existing_broker_url = cls.get_broker_url_from_logs_dir(logs_dir)
        hostname, port = RedisManager.get_hostname_port_from_url(existing_broker_url)
        return RedisManager(hostname=hostname,
                            port=port,
                            redis_bin_base=get_redis_bin_dir())

    @classmethod
    def get_broker_url(cls, assert_if_not_set=False)-> str:
        url = os.environ.get(cls.broker_env_variable, "")
        if assert_if_not_set and not url:
            raise BrokerManagerException('%s env variable has not been set' % cls.broker_env_variable)
        return url


class BrokerManagerException(Exception):
    pass



import os
import argparse
import logging
import subprocess
import time
from psutil import Process, TimeoutExpired

from celery import Celery

from firexapp.celery_manager import CeleryManager
from firexapp.submit.uid import Uid
from firexapp.broker_manager.broker_factory import BrokerFactory, REDIS_BIN_ENV
from firexkit.inspect import get_active
from firexapp.common import qualify_firex_bin, select_env_vars

logger = logging.getLogger(__name__)


def launch_background_shutdown(logs_dir):
    try:
        pid = subprocess.Popen([qualify_firex_bin("firex_shutdown"), "--logs_dir",  logs_dir],
                               close_fds=True, env=select_env_vars([REDIS_BIN_ENV, 'PATH'])).pid
    except Exception as e:
        logger.error("SHUTDOWN PROCESS FAILED TO LAUNCH -- REDIS WILL LEAK.")
        logger.error(e)
        raise
    else:
        try:
            Process(pid).wait(0.1)
        except TimeoutExpired:
            logger.debug("Started background shutdown with pid %s" % pid)
        else:
            logger.error("SHUTDOWN PROCESS FAILED TO RUN -- REDIS WILL LEAK.")


def wait_for_broker_shutdown(broker, timeout=15):
    logger.debug("Waiting for broker to shut down")
    shutdown_wait_time = time.time() + timeout
    while time.time() < shutdown_wait_time:
        if not broker.is_alive():
            break
        time.sleep(0.1)

    if broker.is_alive():
        logger.debug("Warning! Broker was not shut down after %s seconds. FORCE KILLING BROKER." % str(timeout))
        broker.force_kill()
    else:
        logger.debug("Confirmed successful graceful broker shutdown.")


def revoke_active_tasks(celery_app):
    revoked = []
    active = get_active(celery_app=celery_app)
    if active:
        for host in active.values():
            if host:
                logger.info('Revoking lingering tasks...')
            for task in host:
                revoked.append(task["id"])
                logger.warning("Revoking " + task['name'])
                celery_app.control.revoke(task_id=task["id"], terminate=True)
    return revoked


def init():
    parser = argparse.ArgumentParser()
    parser.add_argument("--logs_dir", help="Logs directory for the firexapp run to shutdown.",
                        required=True)
    args = parser.parse_args()
    logs_dir = args.logs_dir

    log_file = os.path.join(logs_dir, Uid.debug_dirname, 'shutdown.log')
    logging.basicConfig(filename=log_file, level=logging.DEBUG, filemode='w',
                        format='[%(asctime)s %(levelname)s] %(message)s', datefmt="%Y-%m-%d %H:%M:%S")

    return logs_dir


def shutdown_run(logs_dir):
    broker = BrokerFactory.broker_manager_from_logs_dir(logs_dir)
    celery_manager = CeleryManager(logs_dir=logs_dir, broker=broker)
    celery_app = Celery(broker=broker.broker_url)

    try:
        #
        # Note: get_active can hang in celery/kombu library if broker is down.
        # Checking for broker.is_alive() is an attempt to prevent that, but note
        # the broker can theoretically die between that call and the get_active() call.
        if broker.is_alive() and get_active(inspect_retry_timeout=2, celery_app=celery_app):
            revoke_active_tasks(celery_app)

            logger.info("Found active Celery; sending Celery shutdown.")
            celery_app.control.shutdown()

            celery_shutdown_wait = 15
            celery_shutdown_success = celery_manager.wait_for_shutdown(celery_shutdown_wait)
            if not celery_shutdown_success:
                logger.warning("Celery not shutdown after %d secs, force killing instead." % celery_shutdown_wait)
                celery_manager.shutdown()
            else:
                logger.debug("Confirmed Celery shutdown successfully.")
        elif celery_manager.find_all_procs():
            logger.info("Celery not active, but found celery processes to force shutdown.")
            celery_manager.shutdown()
        else:
            logger.info("No active Celery processes.")
    finally:
        if broker.is_alive():
            logger.info("Broker is alive; sending redis shutdown.")
            # broker will be shut down by celery if active
            broker.shutdown()
            wait_for_broker_shutdown(broker)
        else:
            logger.info("No active Broker.")

    logger.info("Shutdown of %s complete." % logs_dir)


def main():
    logs_dir = init()
    try:
        shutdown_run(logs_dir)
    except Exception as e:
        logger.exception(e)



import sys
import logging
CONSOLE_LOGGING_FORMATTER = '[%(asctime)s] %(message)s'

console_stdout = None


class DistlibWarningsFilter(logging.Filter):
    def filter(self, record):
        pathname = record.pathname
        return not pathname.endswith('distlib/metadata.py') and not pathname.endswith('distlib/database.py')


class RetryFilter(logging.Filter):
    def filter(self, record):
        return not record.getMessage().startswith('Retry in')


def setup_console_logging(module=None):
    global console_stdout

    formatter = logging.Formatter(CONSOLE_LOGGING_FORMATTER, "%H:%M:%S")
    if module == "__main__":
        # For program entry point, use root logger
        module_logger = logging.getLogger()
        # noinspection PyUnresolvedReferences
        module_logger.setLevel(logging.DEBUG)
    else:
        # For submodules, create sub-loggers
        module_logger = logging.getLogger(module)
        module_logger.setLevel(logging.NOTSET)
        return module_logger

    class LogLevelFilter(logging.Filter):
        """Filters (lets through) all messages with level < LEVEL"""

        def __init__(self, level):
            self.level = level
            super(LogLevelFilter, self).__init__()

        def filter(self, record):
            # "<" instead of "<=": since logger.setLevel is inclusive, this should
            # be exclusive
            return record.levelno < self.level

    console_stdout = logging.StreamHandler(sys.stdout)
    console_stdout.setLevel(logging.INFO)
    console_stdout.setFormatter(formatter)
    console_stdout.addFilter(LogLevelFilter(logging.ERROR))
    console_stdout.addFilter(DistlibWarningsFilter())

    console_stderr = logging.StreamHandler()
    console_stderr.setLevel(logging.ERROR)
    console_stderr.setFormatter(formatter)
    console_stderr.addFilter(RetryFilter())

    module_logger.addHandler(console_stdout)
    module_logger.addHandler(console_stderr)
    return module_logger


def set_console_log_level(log_level):
    global console_stdout
    console_stdout.setLevel(log_level)


def add_filter_to_console(log_filter):
    assert isinstance(log_filter, logging.Filter)
    console_stdout.addFilter(log_filter)
    set_console_log_level(logging.DEBUG)
